# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command group for spanner samples."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope import base


class Samples(base.Group):
  """Cloud Spanner sample apps.

  Each Cloud Spanner sample application includes a backend gRPC service
  backed by a Cloud Spanner database and a workload script that generates
  service traffic.

  These sample apps are open source and available at
  https://github.com/GoogleCloudPlatform/cloud-spanner-samples.

  To see a list of available sample apps, run:

      $ gcloud spanner samples list
  """
  pass



# -*- coding: utf-8 -*- #
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Declarative hooks for `gcloud dialogflow agent`."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import base64

from googlecloudsdk.api_lib.storage import storage_util
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core.util import files


def AddSessionPrefix(session):
  project = properties.VALUES.core.project.GetOrFail()
  return 'projects/{}/agent/sessions/{}'.format(project, session)


def SetQueryLanguage(unused_instance_ref, args, request):
  query_input = request.googleCloudDialogflowV2DetectIntentRequest.queryInput
  if args.IsSpecified('query_text'):
    query_input.text.languageCode = args.language
  elif args.IsSpecified('query_audio_file'):
    query_input.audioConfig.languageCode = args.language
  return request


def LogTrainSuccess(unused_response, unused_args):
  log.status.Print('Successfully trained agent.')


def IsBucketUri(path):
  return path.startswith(storage_util.GSUTIL_BUCKET_PREFIX)


def SetAgentUri(unused_instance_ref, args, request):
  dest = args.destination
  if IsBucketUri(dest) and storage_util.ValidateBucketUrl:
    request.googleCloudDialogflowV2ExportAgentRequest = {'agentUri': dest}
  return request


def SaveAgentToFile(response, args):
  dest = args.destination
  if not IsBucketUri(dest):
    props = response.additionalProperties
    agent_content = next(prop for prop in props if prop.key == 'agentContent')
    agent_content_bin = base64.b64decode(agent_content.value.string_value)
    log.WriteToFileOrStdout(dest, agent_content_bin, binary=True)
    if dest != '-':
      log.status.Print('Wrote agent to [{}].'.format(dest))
  return response


def ChooseImportOrRestoreMethod(unused_instance_ref, args):
  if args.replace_all:
    return 'restore'
  return 'import'


def _GetAgentRequestBody(source):
  if source.startswith('gs://'):
    return {'agentUri': source}
  else:
    return {'agentContent': files.ReadBinaryFileContents(source)}


def ModifyImportOrRestoreRequest(unused_instance_ref, args, request):
  body = _GetAgentRequestBody(args.source)

  if args.replace_all:
    request.googleCloudDialogflowV2RestoreAgentRequest = body
  else:
    request.googleCloudDialogflowV2ImportAgentRequest = body

  return request


def LogImportSuccess(response, args):
  path = args.source
  if not args.async_:
    if path != '-':
      log.status.Print('Successfully imported agent from [{}].'.format(path))
    else:
      log.status.Print('Successfully imported agent.')
    if args.replace_all:
      log.status.Print('Replaced all existing resources.')
  return response



# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Command for moving addresses."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.compute import base_classes
from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.compute import flags as compute_flags
from googlecloudsdk.command_lib.compute.addresses import flags


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class Move(base.SilentCommand):
  """Move an address to another project."""

  ADDRESS_ARG = None

  @classmethod
  def Args(cls, parser):
    cls.ADDRESS_ARG = flags.AddressArgument(plural=False)
    cls.ADDRESS_ARG.AddArgument(parser)
    flags.AddMoveArguments(parser)

  def Run(self, args):
    holder = base_classes.ComputeApiHolder(self.ReleaseTrack())
    address_ref = self.ADDRESS_ARG.ResolveAsResource(
        args,
        holder.resources,
        scope_lister=compute_flags.GetDefaultScopeLister(holder.client))
    new_name = args.new_name if args.new_name is not None else address_ref.Name(
    )

    messages = holder.client.messages
    if address_ref.Collection() == 'compute.globalAddresses':
      address_url = 'projects/{}/global/addresses/{}'.format(
          args.target_project, new_name)
      request_msg = messages.ComputeGlobalAddressesMoveRequest(
          address=address_ref.Name(),
          project=address_ref.project,
          globalAddressesMoveRequest=messages.GlobalAddressesMoveRequest(
              description=args.description,
              destinationAddress=address_url,
          ),
      )
      request = (holder.client.apitools_client.globalAddresses, 'Move',
                 request_msg)
    else:
      address_url = 'projects/{}/regions/{}/addresses/{}'.format(
          args.target_project, address_ref.region, new_name)
      request_msg = messages.ComputeAddressesMoveRequest(
          region=address_ref.region,
          address=address_ref.Name(),
          project=address_ref.project,
          regionAddressesMoveRequest=messages.RegionAddressesMoveRequest(
              description=args.description,
              destinationAddress=address_url,
          ),
      )
      request = (holder.client.apitools_client.addresses, 'Move', request_msg)
    return holder.client.MakeRequests([request],
                                      project_override=args.target_project)



# -*- coding: utf-8 -*- #
# Copyright 2019 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for working with dictionaries representing yaml files."""
from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals


def GetOrCreate(obj, path, constructor=dict):
  """Get or create the object by following the field names in the path.

  not exist, create the appropriate value.

  Args:
    obj: A dictionary representing a yaml dictionary
    path: A list of strings representing fields to follow.
    constructor: If the object at the end of the path does not exist, create an
      object using the constructor given.

  Returns:
    An object at found by following the path.
  """
  first, rest = path[0], path[1:]

  if rest:
    if first not in obj:
      obj[first] = dict()
    return GetOrCreate(obj[first], rest, constructor)
  else:
    if first not in obj:
      obj[first] = constructor()
    return obj[first]


def GetAll(obj, path):
  """Given a yaml object, yield all objects found by following a path.

  Given a yaml object, read each field in the path and return the object
  found at the end. If a field has a list value, follow the path for each
  object in the list.

  E.g.
  >>> X = {'A': {'B': [{'C': {'D': 1}}, {'C': {'D': 2}}]}}
  >>> sorted(list(GetAll(X, path=('A', 'B', 'C', 'D'))))
  [1, 2]

  Args:
    obj: A dictionary representing a yaml dictionary
    path: A list of strings representing fields to follow.

  Yields:
    Values that are found by following the given path.
  """
  if not path:
    yield obj
    return

  first, rest = path[0], path[1:]
  if first in obj:
    if isinstance(obj[first], dict):
      for extracted_obj in GetAll(obj[first], rest):
        yield extracted_obj
    elif isinstance(obj[first], list):
      for x in obj[first]:
        for extracted_obj in GetAll(x, rest):
          yield extracted_obj
    else:
      if rest:
        raise ValueError(first + ' is not a dictionary or a list')
      else:
        yield obj[first]



# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Implementation functions for downloads from the Google Cloud Storage API."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from apitools.base.py import http_wrapper as apitools_http_wrapper

from googlecloudsdk.api_lib.storage import errors as cloud_errors
from googlecloudsdk.api_lib.storage import retry_util
from googlecloudsdk.calliope import exceptions as calliope_errors
from googlecloudsdk.core import exceptions as core_exceptions
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core.util import retry

import oauth2client


def _no_op_callback(unused_response, unused_object):
  """Disables Apitools' default print callbacks."""
  pass


def launch(apitools_download,
           start_byte=0,
           end_byte=None,
           additional_headers=None):
  """GCS-specific download implementation.

  Args:
    apitools_download (apitools.transfer.Download): Apitools object for managing
      downloads.
    start_byte (int): Starting point for download (for resumable downloads and
      range requests). Can be set to negative to request a range of bytes
      (python equivalent of [:-3]).
    end_byte (int): Ending byte number, inclusive, for download (for range
      requests). If None, download the rest of the object.
    additional_headers (dict|None): Headers to add to HTTP request.

  Returns:
    Encoding string for object if requested. Otherwise, None.
  """
  if start_byte or end_byte is not None:
    apitools_download.GetRange(
        additional_headers=additional_headers,
        start=start_byte,
        end=end_byte,
        use_chunks=False)
  else:
    apitools_download.StreamMedia(
        additional_headers=additional_headers,
        callback=_no_op_callback,
        finish_callback=_no_op_callback,
        use_chunks=False)
  return apitools_download.encoding


def launch_retriable(download_stream,
                     apitools_download,
                     start_byte=0,
                     end_byte=None,
                     additional_headers=None):
  """Wraps download to make it retriable."""
  # Hack because nonlocal keyword causes Python 2 syntax error.
  progress_state = {'start_byte': start_byte}
  retry_util.set_retry_func(apitools_download)

  def _should_retry_resumable_download(exc_type, exc_value, exc_traceback,
                                       state):
    converted_error, _ = calliope_errors.ConvertKnownError(exc_value)
    if isinstance(exc_value, oauth2client.client.HttpAccessTokenRefreshError):
      if exc_value.status < 500 and exc_value.status != 429:
        # Not server error or too many requests error.
        return False
    elif not (isinstance(converted_error, core_exceptions.NetworkIssueError) or
              isinstance(converted_error, cloud_errors.RetryableApiError)):
      # Not known transient network error.
      return False

    start_byte = download_stream.tell()
    if start_byte > progress_state['start_byte']:
      # We've made progress, so allow a fresh set of retries.
      progress_state['start_byte'] = start_byte
      state.retrial = 0
    log.debug('Retrying download from byte {} after exception: {}.'
              ' Trace: {}'.format(start_byte, exc_type, exc_traceback))

    apitools_http_wrapper.RebuildHttpConnections(apitools_download.bytes_http)
    return True

  def _call_launch():
    return launch(
        apitools_download,
        start_byte=progress_state['start_byte'],
        end_byte=end_byte,
        additional_headers=additional_headers)

  # Convert seconds to miliseconds by multiplying by 1000.
  return retry.Retryer(
      max_retrials=properties.VALUES.storage.max_retries.GetInt(),
      wait_ceiling_ms=properties.VALUES.storage.max_retry_delay.GetInt() * 1000,
      exponential_sleep_multiplier=(
          properties.VALUES.storage.exponential_sleep_multiplier.GetInt()
      )).RetryOnException(
          _call_launch,
          sleep_ms=properties.VALUES.storage.base_retry_delay.GetInt() * 1000,
          should_retry_if=_should_retry_resumable_download)



# -*- coding: utf-8 -*- #
# Copyright 2016 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flags and helpers for the datastore related commands."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope import arg_parsers


def AddEntityFilterFlags(parser):
  """Adds flags for entity filters to the given parser."""
  parser.add_argument(
      '--kinds',
      metavar='KIND',
      type=arg_parsers.ArgList(),
      help="""
      A list specifying what kinds will be included in the operation. When
      omitted, all Kinds are included. For example, to operate on only the
      'Customer' and 'Order' Kinds:

        $ {command} --kinds='Customer','Order'
      """)

  parser.add_argument(
      '--namespaces',
      metavar='NAMESPACE',
      type=arg_parsers.ArgList(),
      help="""
      A list specifying what namespaces will be included in the operation.
      When omitted, all namespaces are included in the operation,
      including the default namespace. To specify that *only* the default
      namespace should be operated on, use the special symbol '(default)'.
      For example, to operate on entities from both the 'customers' and default
      namespaces:

        $ {command} --namespaces='(default)','customers'
      """)


def AddLabelsFlag(parser):
  """Adds a --operation-labels flag to the given parser."""
  parser.add_argument(
      '--operation-labels',
      metavar='OPERATION_LABEL',
      type=arg_parsers.ArgDict(),
      help="""
      A string:string map of custom labels to associate with this operation.
      For example:

        $ {command} --operation-labels=comment='customer orders','sales rep'=pending
      """)



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Removes an invoker binding from the IAM policy of a Google Cloud Function."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope import base
from googlecloudsdk.command_lib.functions import flags
from googlecloudsdk.command_lib.functions.v2.remove_invoker_policy_binding import command
from googlecloudsdk.command_lib.iam import iam_util


_DETAILED_HELP = {
    'DESCRIPTION':
        '{description}',
    'EXAMPLES':
        """\
          To remove the invoker policy binding for `FUNCTION-1` from role
          `ROLE-1` for member `MEMBER-1` run:

            $ {command} FUNCTION-1 --member=MEMBER-1
          """,
}


def _CommonArgs(parser, track):
  """Registers flags for this command."""
  del track
  flags.AddFunctionResourceArg(parser, 'to remove the invoker binding from')
  iam_util.AddMemberFlag(parser, 'to remove from the IAM policy', False)


@base.ReleaseTracks(base.ReleaseTrack.GA)
class RemoveInvokerPolicyBinding(base.Command):
  """Removes an invoker binding from the IAM policy of a Google Cloud Function.

  This command applies to Cloud Functions V2 only.
  """

  detailed_help = _DETAILED_HELP

  @staticmethod
  def Args(parser):
    """Registers flags for this command."""
    _CommonArgs(parser, base.ReleaseTrack.GA)

  def Run(self, args):
    """Runs the command.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      The updated IAM policy for the service.
    """
    return command.Run(args, self.ReleaseTrack())


@base.ReleaseTracks(base.ReleaseTrack.BETA)
class RemoveInvokerPolicyBindingBeta(base.Command):
  """Removes an invoker binding from the IAM policy of a Google Cloud Function.

  This command applies to Cloud Functions V2 only.
  """
  detailed_help = _DETAILED_HELP

  @staticmethod
  def Args(parser):
    """Registers flags for this command."""
    _CommonArgs(parser, base.ReleaseTrack.BETA)

  def Run(self, args):
    """Runs the command.

    Args:
      args: an argparse namespace. All the arguments that were provided to this
        command invocation.

    Returns:
      The updated IAM policy for the service.
    """
    return command.Run(args, self.ReleaseTrack())


@base.ReleaseTracks(base.ReleaseTrack.ALPHA)
class RemoveInvokerPolicyBindingAlpha(RemoveInvokerPolicyBindingBeta):
  """Removes an invoker binding from the IAM policy of a Google Cloud Function.

  This command applies to Cloud Functions V2 only.
  """

  @staticmethod
  def Args(parser):
    """Registers flags for this command."""
    _CommonArgs(parser, base.ReleaseTrack.ALPHA)



# -*- coding: utf-8 -*- #
# Copyright 2017 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities Cloud IoT registries API."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from apitools.base.py import list_pager
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.core import exceptions


class NoFieldsSpecifiedError(exceptions.Error):
  """Error when no fields were specified for a Patch operation."""


def GetClientInstance(no_http=False):
  return apis.GetClientInstance('cloudiot', 'v1', no_http=no_http)


def GetMessagesModule(client=None):
  client = client or GetClientInstance()
  return client.MESSAGES_MODULE


class _DeviceRegistryUpdateSetting(object):
  """Small value class holding data for updating a device registry."""

  def __init__(self, field_name, update_mask, value):
    self.field_name = field_name
    self.update_mask = update_mask
    self.value = value


class RegistriesClient(object):
  """Client for registries service in the Cloud IoT API."""

  def __init__(self, client=None, messages=None):
    self.client = client or GetClientInstance()
    self.messages = messages or GetMessagesModule(client)
    self._service = self.client.projects_locations_registries

  @property
  def mqtt_config_enum(self):
    return self.messages.MqttConfig.MqttEnabledStateValueValuesEnum

  @property
  def http_config_enum(self):
    return self.messages.HttpConfig.HttpEnabledStateValueValuesEnum

  def Create(self, parent_ref, registry_id, credentials=None,
             event_notification_configs=None, state_pubsub_topic=None,
             mqtt_enabled_state=None, http_enabled_state=None):
    """Creates a DeviceRegistry.

    Args:
      parent_ref: a Resource reference to a cloudiot.projects.locations
        resource for the parent of this registry.
      registry_id: str, the name of the resource to create.
      credentials: List of RegistryCredentials or None, credentials for the
        DeviceRegistry.
      event_notification_configs: List of EventNotificationConfigs or None,
        configs for forwarding telemetry events for the Registry.
      state_pubsub_topic: an optional Resource reference to a
        pubsub.projects.topics. The pubsub topic for state notifications on this
        device registry.
      mqtt_enabled_state: MqttEnabledStateValueValuesEnum, the state of MQTT for
        the registry.
      http_enabled_state: HttpEnabledStateValueValuesEnum, the state of HTTP for
        the registry.

    Returns:
      DeviceRegistry: the created registry.
    """
    if state_pubsub_topic:
      # This is a repeated field, only the first entry is used though.
      state_notification_config = self.messages.StateNotificationConfig(
          pubsubTopicName=state_pubsub_topic.RelativeName())
    else:
      state_notification_config = None

    if mqtt_enabled_state:
      mqtt_config = self.messages.MqttConfig(
          mqttEnabledState=mqtt_enabled_state)
    else:
      mqtt_config = None

    if http_enabled_state:
      http_config = self.messages.HttpConfig(
          httpEnabledState=http_enabled_state)
    else:
      http_config = None

    create_req = self.messages.CloudiotProjectsLocationsRegistriesCreateRequest(
        parent=parent_ref.RelativeName(),
        deviceRegistry=self.messages.DeviceRegistry(
            id=registry_id,
            credentials=credentials or [],
            eventNotificationConfigs=event_notification_configs or [],
            stateNotificationConfig=state_notification_config,
            mqttConfig=mqtt_config,
            httpConfig=http_config))

    return self._service.Create(create_req)

  def Delete(self, registry_ref):
    delete_req = self.messages.CloudiotProjectsLocationsRegistriesDeleteRequest(
        name=registry_ref.RelativeName())
    return self._service.Delete(delete_req)

  def Get(self, registry_ref):
    get_req = self.messages.CloudiotProjectsLocationsRegistriesGetRequest(
        name=registry_ref.RelativeName())
    return self._service.Get(get_req)

  def List(self, parent_ref, limit=None, page_size=100):
    """List the device registries in a given location.

    Args:
      parent_ref: a Resource reference to a cloudiot.projects.locations
        resource to list devices for.
      limit: int, the total number of results to return from the API.
      page_size: int, the number of results in each batch from the API.

    Returns:
      A generator of the device registries in the location.
    """
    list_req = self.messages.CloudiotProjectsLocationsRegistriesListRequest(
        parent=parent_ref.RelativeName())
    return list_pager.YieldFromList(
        self._service, list_req, batch_size=page_size, limit=limit,
        field='deviceRegistries', batch_size_attribute='pageSize')

  def Patch(self,
            registry_ref,
            credentials=None,
            event_notification_configs=None,
            state_pubsub_topic=None,
            mqtt_enabled_state=None,
            http_enabled_state=None,
            log_level=None):
    """Updates a DeviceRegistry.

    Any fields not specified will not be updated; at least one field must be
    specified.

    Args:
      registry_ref: a Resource reference to a
        cloudiot.projects.locations.registries resource.
      credentials: List of RegistryCredentials or None, credentials for the
        DeviceRegistry.
      event_notification_configs: List of EventNotificationConfigs or None,
        configs for forwarding telemetry events for the Registry.
      state_pubsub_topic: an optional Resource reference to a
        pubsub.projects.topics. The pubsub topic for state notifications on this
        device registry.
      mqtt_enabled_state: MqttEnabledStateValueValuesEnum, the state of MQTT for
        the registry.
      http_enabled_state: HttpConfigStateValuEnabledsEnum, the state of HTTP for
        the registry.
      log_level: LogLevelValueValuesEnum, the default logging verbosity for the
        devices in the registry.

    Returns:
      DeviceRegistry: the created registry.

    Raises:
      NoFieldsSpecifiedError: if no fields were specified.
    """
    registry = self.messages.DeviceRegistry()

    if state_pubsub_topic:
      # This is a repeated field, only the first entry is used though.
      state_notification_config = self.messages.StateNotificationConfig(
          pubsubTopicName=state_pubsub_topic.RelativeName())
    else:
      state_notification_config = None

    if mqtt_enabled_state:
      mqtt_config = self.messages.MqttConfig(
          mqttEnabledState=mqtt_enabled_state)
    else:
      mqtt_config = None

    if http_enabled_state:
      http_config = self.messages.HttpConfig(
          httpEnabledState=http_enabled_state)
    else:
      http_config = None

    device_registry_update_settings = [
        _DeviceRegistryUpdateSetting('credentials', 'credentials', credentials),
        _DeviceRegistryUpdateSetting('eventNotificationConfigs',
                                     'event_notification_configs',
                                     event_notification_configs),
        _DeviceRegistryUpdateSetting(
            'stateNotificationConfig',
            'state_notification_config.pubsub_topic_name',
            state_notification_config),
        _DeviceRegistryUpdateSetting(
            'mqttConfig', 'mqtt_config.mqtt_enabled_state', mqtt_config),
        _DeviceRegistryUpdateSetting(
            'httpConfig', 'http_config.http_enabled_state', http_config),
        _DeviceRegistryUpdateSetting('logLevel', 'logLevel', log_level)
    ]
    update_mask = []
    for update_setting in device_registry_update_settings:
      if update_setting.value is not None:
        setattr(registry, update_setting.field_name, update_setting.value)
        update_mask.append(update_setting.update_mask)
    if not update_mask:
      raise NoFieldsSpecifiedError('Must specify at least one field to update.')
    patch_req = self.messages.CloudiotProjectsLocationsRegistriesPatchRequest(
        deviceRegistry=registry,
        name=registry_ref.RelativeName(),
        updateMask=','.join(update_mask))

    return self._service.Patch(patch_req)

  def SetIamPolicy(self, registry_ref, set_iam_policy_request):
    """Sets an IAM policy on a DeviceRegistry.

    Args:
      registry_ref: a Resource reference to a
        cloudiot.projects.locations.registries resource.
      set_iam_policy_request: A SetIamPolicyRequest which contains the Policy to
        add.

    Returns:
      Policy: the added policy.
    """
    set_req = (
        self.messages.CloudiotProjectsLocationsRegistriesSetIamPolicyRequest(
            resource=registry_ref.RelativeName(),
            setIamPolicyRequest=set_iam_policy_request))
    return self._service.SetIamPolicy(set_req)

  def GetIamPolicy(self, registry_ref):
    """Gets the IAM policy for a DeviceRegistry.

    Args:
      registry_ref: a Resource reference to a
        cloudiot.projects.locations.registries resource.

    Returns:
      Policy: the policy for the device registry.
    """
    get_req = (
        self.messages.CloudiotProjectsLocationsRegistriesGetIamPolicyRequest(
            resource=registry_ref.RelativeName()))
    return self._service.GetIamPolicy(get_req)
